import webbrowser
from commoner import *
from .main import *
from .utils import *


import string

from typing import Dict
from nicegui import ui


def gui():
    """
    Start the GUI.
    """
    ui.colors(primary="#299fbb")

    WIDTH = "w-2/3"

    def search():
        def result(molecule):
            bottom_content.clear()
            if (
                molecule not in string.whitespace
                and molecule != ""
                and molecule is not None
            ):
                molecule = Formula(molecule)
                with bottom_content:
                    result_box = ui.card().classes(WIDTH)
                    with result_box:
                        if molecule.name != "Unknown":
                            ui.label("Name").classes("font-bold mt-2")
                            ui.label(f"{molecule.name}")
                            ui.separator()
                        ui.label("Molecular formula").classes("font-bold mt-2")
                        ui.label(f"{molecule.mass} g/mol")
                        if len(molecule.elements) > 1:
                            ui.separator()
                            ui.label("Percent composition").classes("font-bold mt-2")
                            composition = get_percent_composition(molecule, False)
                            for element in composition.items():
                                ui.label(
                                    f"{element[0]}: {round(element[1], 2)}% ({round(element[1], 6)}%, {Element(element[0]).mass} g/mol)"
                                )
                        else:
                            molecule = Element(list(molecule.elements.keys())[0])
                            if molecule.data["Metal"] == "yes":
                                ui.separator()
                                ui.label("Type: Metal").classes("font-bold mt-2")
                            elif molecule.data["Metalloid"] == "yes":
                                ui.separator()
                                ui.label("Type: Metalloid").classes("font-bold mt-2")
                            elif molecule.data["Nonmetal"] == "yes":
                                ui.separator()
                                ui.label("Type: Nonmetal").classes("font-bold mt-2")

        top_content.clear()
        with top_content:
            input_box = ui.card().classes(WIDTH)
        with input_box:
            with ui.card().style("border: 1px solid var(--q-info);"):
                ui.icon("info", color="blue")
                ui.label(
                    "Enter a formula or element and press submit to search for it. This will display the name, molar mass, percent composition, and type of the molecule."
                )

            def submit(value):
                if value is not None:
                    result(value)

            molecule_input = ui.input(
                label="Molecule",
                placeholder="Ex. Fe or H2O",
                validation={"Invalid molecule": lambda x: isformula(x)},
            )
            with ui.row():
                ui.button("Submit", on_click=lambda: submit(molecule_input.value))
                ui.button(
                    "Clear",
                    on_click=lambda: (
                        molecule_input.set_value(""),
                        bottom_content.clear(),
                    ),
                )

    def conversions():
        bottom_content.clear()
        top_content.clear()

        def clear_all():
            input_a.set_value("")
            input_b.set_value("")
            bottom_content.clear()

        def result(value1, value2, operation):
            bottom_content.clear()
            with bottom_content:
                result_box = ui.card().classes(WIDTH)
            if operation == 1:
                with result_box:
                    ui.label(
                        f"{value1} g {value2} = {round(mass_to_moles(Formula(value2), value1), 2)} mol {value2}"
                    )
            elif operation == 2:
                with result_box:
                    ui.label(
                        f"{value1} g {value2} = {round(mass_to_atoms(Formula(value2), value1), 2)} atoms {value2}"
                    )
            elif operation == 3:
                with result_box:
                    ui.label(
                        f"{value1} g {value2} = {round(mass_to_particles(Formula(value2), value1), 2)} particles {value2}"
                    )
            elif operation == 4:
                with result_box:
                    ui.label(
                        f"{value1} mol {value2} = {round(moles_to_mass(Formula(value2), value1), 2)} g {value2}"
                    )
            elif operation == 5:
                with result_box:
                    ui.label(
                        f"{value1} mol {value2} = {round(moles_to_atoms(Formula(value2), value1), 2)} atoms {value2}"
                    )
            elif operation == 6:
                with result_box:
                    ui.label(
                        f"{value1} atoms {value2} = {round(atoms_to_mass(Formula(value2), value1), 2)} g {value2}"
                    )
            elif operation == 7:
                with result_box:
                    ui.label(
                        f"{value1} atoms {value2} = {round(atoms_to_moles(Formula(value2), value1), 2)} mol {value2}"
                    )
            elif operation == 8:
                with result_box:
                    ui.label(
                        f"{value1} particles {value2} = {round(particles_to_mass(Formula(value2), value1), 2)} g {value2}"
                    )

        top_content.clear()
        with top_content:
            input_box = ui.card().classes("w-2/3")
        with input_box:
            with ui.card().style("border: 1px solid var(--q-negative);"):
                ui.icon("warning", color="red")
                ui.label(
                    "This feature is still in development and may not work properly. Confirm calculations are within expected range."
                )

            def submit(value, value2, value3):
                if all([value, value2, value3]):
                    result(value, value2, value3)

            operations = {
                1: "Mass to moles",
                2: "Mass to atoms",
                3: "Mass to particles",
                4: "Moles to mass",
                5: "Moles to atoms",
                6: "Atoms to mass",
                7: "Atoms to moles",
                8: "Particles to mass",
            }

            input_a = ui.number(
                label="Number",
                placeholder="Ex. 10",
            )
            input_b = ui.input(
                label="Element",
                placeholder="Ex. H",
                validation={"Not an element": lambda x: isformula(x)},
            )
            input_c = ui.select(label="Operation", options=operations, value=1)
            with ui.row():
                ui.button(
                    "Submit",
                    on_click=lambda: submit(
                        input_a.value, input_b.value, input_c.value
                    ),
                )
                ui.button("Clear", on_click=lambda: clear_all())

    def periodic_table():
        def result(data):
            bottom_content.clear()
            if (
                molecule not in string.whitespace
                and molecule != ""
                and molecule is not None
            ):
                molecule = Formula(molecule)
                with bottom_content:
                    result_box = ui.card().classes(WIDTH)
                    with result_box:
                        if molecule.name != "Unknown":
                            ui.label(f"{molecule.name}")
                        ui.label(f"{molecule.mass} g/mol")

        top_content.clear()
        bottom_content.clear()
        top_content.clear()
        with top_content:
            input_box = ui.card().classes(WIDTH)
        with input_box:
            def submit(value):
                if value is not None:
                    result(value)

            data = PERIODIC_TABLE
            columnDefs = [{"headerName": col, "field": col} for col in data.columns[:7]]
            rowData = data.to_dict("records")
            grid = ui.aggrid(
                {
                    "columnDefs": columnDefs,
                    "rowData": rowData,
                    "rowSelection": "single",
                }
            )

    def switch(menu, menu_item, text, func, **kwargs):
        menu.close()
        menu_item.set_text(text)
        func(**kwargs)

    with ui.header().classes(replace="row justify-between h-8"):
        ui.label("Chemic").classes("text-2xl font-bold ml-4 mr-2")
        ui.button("GitHub", on_click=lambda: webbrowser.open("https://github.com/uncenter/chemic"))

    
    with ui.row():
        with ui.menu() as menu:
            ui.menu_item(
                "Search",
                on_click=lambda: switch(menu, menu_toggle, "Search", search),
            )
            ui.menu_item(
                "Periodic table",
                on_click=lambda: switch(
                    menu, menu_toggle, "Periodic table", periodic_table
                ),
            )
            ui.separator()
            ui.menu_item(
                "Conversions",
                lambda: switch(menu, menu_toggle, "Conversions", conversions),
            )
        menu_toggle = ui.button("Select", on_click=menu.open).classes("mb-2")
    top_content = ui.row().classes("mt-2")
    bottom_content = ui.row().classes("mt-2")

    ui.run()


def cli():
    """
    Start the CLI (deprecated, use `gui()` instead).
    """

    def search():
        Console.clear()
        string = input("Enter an element/formula: ")
        if iselement(string):
            molecule = Element(string)
            molecule.display()
        elif isformula(string):
            molecule = Formula(string)
        else:
            Shout.error("Invalid input")
            search()
        if input("View more details? (y/N) ").lower() == "y":
            molecule.details()
            if input("View another element? (Y/n) ").lower() in ["y", ""]:
                search()
        else:
            if input("View another element? (Y/n) ").lower() in ["y", ""]:
                search()
            else:
                menu()

    def details():
        Console.clear()
        string = input("Enter an element/formula: ")
        if iselement(string):
            molecule = Element(string)
            molecule.details()
        elif isformula(string):
            molecule = Formula(string)
            molecule.details()
        else:
            Shout.error("Invalid input")
            details()
        if input("View another element? (Y/n) ").lower() in ["y", ""]:
            details()
        else:
            menu()

    def conversions():
        Console.clear()
        choice = create_menu(
            "Mass to moles",
            "Moles to mass",
            "Atoms to moles",
            "Moles to atoms",
            "Particles to mass",
            "Mass to particles",
            "Back",
        )
        if choice == "1":
            mass = float(input("Enter the mass: "))
            element = input("Enter the element: ")
            print(
                f"{mass} g {element} = {round(mass_to_moles(Element(element), mass))} mol {element}"
            )
        elif choice == "2":
            moles = float(input("Enter the moles: "))
            element = input("Enter the element: ")
            print(
                f"{moles} mol {element} = {moles_to_mass(Element(element), moles)} g {element}"
            )
        elif choice == "3":
            atoms = float(input("Enter the atoms: "))
            element = input("Enter the element: ")
            print(
                f"{atoms} {element} = {atoms_to_moles(Element(element), atoms)} mol {element}"
            )
        elif choice == "4":
            moles = float(input("Enter the moles: "))
            element = input("Enter the element: ")
            print(
                f"{moles} mol {element} = {moles_to_atoms(Element(element), moles)} {element}"
            )
        elif choice == "5":
            particles = float(input("Enter the particles: "))
            element = input("Enter the element: ")
            print(
                f"{particles} {element} = {particles_to_mass(Element(element), particles)} g {element}"
            )
        elif choice == "6":
            mass = float(input("Enter the mass: "))
            element = input("Enter the element: ")
            print(
                f"{mass} g {element} = {mass_to_particles(Element(element), mass)} {element}"
            )
        elif choice == "7":
            menu()
        else:
            Shout.error("Invalid choice")
            conversions()
        if input("Again? (Y/n) ").lower() in ["y", ""]:
            conversions()
        else:
            menu()

    def percent_composition():
        Console.clear()
        formula = input("Enter the formula: ")
        if isformula(formula):
            molecule = Formula(formula)
            comp = get_percent_composition(molecule, True)
            for element in comp:
                print(f"{element}: {comp[element]}%")
                if input("Again? (Y/n) ").lower() in ["y", ""]:
                    percent_composition()
                else:
                    menu()
        else:
            Shout.error("Invalid formula")
            percent_composition()

    def empirical_molecular():
        Console.clear()
        choice = create_menu(
            "Calculate empirical formula", "Calculate molecular formula", "Back"
        )
        if choice == "1":
            num_elements = int(input("Enter the number of elements: "))
            for i in range(num_elements):
                element = input(f"Enter the element {i+1}: ")
                mass = float(input(f"Enter the mass of {element}: "))
                if iselement(element):
                    molecule = Element(element)
                    print(f"Empirical formula: {get_empirical_formula(molecule, mass)}")
                else:
                    Shout.error("Invalid element")
                    empirical_molecular()
        elif choice == "2":
            formula = input("Enter the empirical formula: ")
            mass = float(input("Enter the mass: "))
            if isformula(formula):
                print(f"Molecular formula: {get_molecular_formula(formula, mass)}")
            else:
                Shout.error("Invalid formula")
                empirical_molecular()
        elif choice == "3":
            menu()
        else:
            Shout.error("Invalid choice")
            empirical_molecular()
        if input("Again? (Y/n) ").lower() in ["y", ""]:
            empirical_molecular()

    def menu():
        choice = create_menu(
            "Quick search",
            "View details",
            "Conversions and calculations",
            "Percent composition",
            "Empirical formula and molecular formula",
            "Exit",
        )
        if choice == "1":
            search()
        elif choice == "2":
            details()
        elif choice == "3":
            conversions()
        elif choice == "4":
            percent_composition()
        elif choice == "5":
            empirical_molecular()
        elif choice == "6":
            exit()
        else:
            Shout.error("Invalid choice")
            menu()

    Shout.welcome("Chemic", "uncenter", version="0.9", pause=True)
    Shout.warning("This CLI is deprecated, use `gui()` instead.")
    menu()


from commoner import reverse, replace_all, Shout, Chalk


def validate_cas_number(cas_num, verbose=True):
    if not isinstance(cas_num, str):
        if verbose:
            Shout.error("CAS number is not a string")
        return False
    if len(replace_all(cas_num, "-", "")) > 10:
        if verbose:
            Shout.error("CAS number is too long")
        return False
    cas_num = {
        "full": cas_num,
        "clean": replace_all(cas_num, "-", ""),
        "parts": cas_num.split("-"),
    }
    if len(cas_num["parts"]) != 3:
        if verbose:
            Shout.error("CAS number is not formatted correctly [x(2-7)-x(2)-x(1)]")
        return False
    check_sum = 0
    check_sum_num = reverse(cas_num["parts"][0] + cas_num["parts"][1])
    for i in range(0, len(check_sum_num)):
        check_sum += int(check_sum_num[i]) * (i + 1)
    check_sum = check_sum % 10
    if check_sum != int(cas_num["parts"][2]):
        if verbose:
            Shout.error(
                f"Check sum does not match check digit [{check_sum} != {cas_num['parts'][2]}]"
            )
        return False
    if verbose:
        Shout.success("CAS number is valid")
    return True


def parse_formula(formula):
    element_dict = {}
    looking_for_num = False
    skip = False
    element = ""
    for i in range(len(formula)):
        if skip:
            skip = False
            continue
        elif formula[i] == "(":
            if looking_for_num:
                element_dict[element] = 1
                looking_for_num = False
            for j in range(i, len(formula)):
                if formula[j] == ")":
                    index = j
                    break
            result = parse_formula(formula[i + 1 : index])
            multiplier = None
            for k in range(index + 1, len(formula)):
                if formula[k].isdigit():
                    multiplier = int(formula[index + 1 : k + 1])
                else:
                    break
            if multiplier == None:
                multiplier = 1
            for sub_element, sub_num_atoms in result.items():
                if sub_element in element_dict:
                    element_dict[sub_element] += sub_num_atoms * (multiplier)
                else:
                    element_dict[sub_element] = sub_num_atoms * (multiplier)
            formula = formula.replace(formula[: index + (1 + len(str(multiplier)))], "")
            sub_element_dict = parse_formula(formula)
            for sub_element, sub_num_atoms in sub_element_dict.items():
                if sub_element in element_dict:
                    element_dict[sub_element] += sub_num_atoms
                else:
                    element_dict[sub_element] = sub_num_atoms
            return element_dict
        if formula[i].isupper():
            if looking_for_num:
                element_dict[element] = 1
                looking_for_num = False
            element = formula[i]
            num_atoms = 1
            looking_for_num = True
        elif formula[i].islower():
            if i == 0 or not formula[i - 1].isupper():
                if formula[i - 1].islower():
                    element_dict[element] = 1
                element = formula[i].upper()
                num_atoms = 1
            else:
                element += formula[i]
            looking_for_num = True
        elif formula[i].isdigit():
            if i < len(formula) - 1 and formula[i + 1].isdigit():
                num_atoms = int(formula[i : i + 2])
                skip = True
            else:
                num_atoms = int(formula[i])
            if element in element_dict:
                element_dict[element] += num_atoms
            else:
                element_dict[element] = num_atoms
            looking_for_num = False
        else:
            return
    if looking_for_num:
        element_dict[element] = num_atoms
    new_dict = {}
    for element, num_atoms in element_dict.items():
        if element != "":
            new_dict[element] = num_atoms
    return new_dict


def create_menu(*args, cursor="> ", prompt="Select an option: ", color="cyan"):
    print(prompt)
    for i in range(len(args)):
        print(f"{i + 1}. {args[i]}")
    while True:
        try:
            Chalk.set(color)
            choice = input(cursor)
            Chalk.reset()
            if choice in "".join(args) and not choice.isnumeric() and len(choice) > 1:
                if sum([choice in i for i in args]) > 1:
                    raise ValueError
                for i in range(len(args)):
                    if choice in args[i]:
                        return str(i + 1)
            if int(choice) > len(args) or int(choice) < 1:
                raise ValueError
            return str(choice)
        except ValueError:
            print("Invalid choice")



from commoner import *
from .utils import validate_cas_number, parse_formula

import pandas as pd
import pydash
import importlib.resources


PERIODIC_TABLE = pd.read_csv(
    importlib.resources.open_text("chemic", "periodic_table.csv")
)
FORMULA_TABLE = pd.read_csv(
    importlib.resources.open_text("chemic", "common_formulas.csv")
)
AVOGADRO = 6.02214076e23
L_STP = 22.4


def reconstruct_formula(formula):
    """
    Reconstruct a formula from a dictionary.

    Args:
        formula (dict): The formula to reconstruct.

    Returns:
        str: The reconstructed formula.
    """
    reconstruct_formula = []
    for element in formula:
        if formula[element] == 1:
            reconstruct_formula.append(element)
        else:
            reconstruct_formula.append(f"{element}{formula[element]}")
    return "".join(reconstruct_formula)


def get_molar_mass(molecule):
    """
    Get the molar mass of a molecule.

    Args:
        molecule (Formula, Element): The molecule to get the molar mass of.

    Returns:
        float: The molar mass of the molecule.
    """
    molar_mass = 0
    if isinstance(molecule, Formula):
        for element in molecule.elements:
            molar_mass += (molecule.elements[element]) * Element(element).mass
    elif isinstance(molecule, Element):
        return molecule.mass
    else:
        return None
    return molar_mass


def get_formula_name(formula, verbose=False):
    """
    Get the name of a molecule from its formula.

    Args:
        formula (str): The formula of the molecule.
        verbose (bool, optional): Whether to print the process. Defaults to False.

    Returns:
        str: The name of the molecule.
    """
    if type(formula) != str:
        return None
    if validate_cas_number(formula, False):
        if verbose:
            Shout.info("CAS number detected")
        if formula in FORMULA_TABLE["CAS"].values:
            return FORMULA_TABLE[FORMULA_TABLE["CAS"] == formula]["Names"].values[0]
        else:
            return None
    else:
        if verbose:
            Shout.info("Formula detected")
        if formula in FORMULA_TABLE["Formula"].values:
            return (
                FORMULA_TABLE[FORMULA_TABLE["Formula"] == formula]["Names"]
                .values[0]
                .split("\n")[0]
            )
        else:
            return None


class Element:
    """
    A class to represent an element.

    Args:
        attribute (str, int): The attribute to use to initialize the element.

    Attributes:
        data (dict): The data of the element.
        symbol (str): The symbol of the element.
        name (str): The name of the element.
        number (int): The atomic number of the element.
        mass (float): The atomic mass of the element.
        count (int): The number of atoms of the element in a molecule.

    Methods:
        display: Display the element in a "periodic table"-like format.
        details: Display the details of the element.

    Examples:
        >>> element = Element("H")
        >>> element.display()
        ┌────────────────────┐
        │1.008              1│
        │                    │
        │         H          │
        │                    │
        │      Hydrogen      │
        └────────────────────┘
        >>> print(Element("H").details())
        Atomic Number: 1
        Element: Hydrogen
        Symbol: H
        Atomic Mass: 1.008
        Number Of Neutrons: 0
        Number Of Protons: 1
        Number Of Electrons: 1
        ...
    """

    def __init__(self, attribute):
        if type(attribute) == str:
            if attribute.isdigit():
                attribute = int(attribute)
            else:
                attribute = attribute.capitalize()
        attr_types = ["Symbol", "Element", "AtomicNumber", "AtomicMass"]
        for attr_type in attr_types:
            if attribute in PERIODIC_TABLE[attr_type].values:
                self.data = PERIODIC_TABLE[
                    PERIODIC_TABLE[attr_type] == attribute
                ].to_dict("records")[0]
                self.symbol = self.data["Symbol"]
                self.name = self.data["Element"]
                self.number = self.data["AtomicNumber"]
                self.mass = self.data["AtomicMass"]
                self.count = 1
                break
        else:
            return None

    def display(self):
        """
        Display the element in a "periodic table"-like format.
        """
        # References:
        # https://docs.python.org/3/library/string.html#format-specification-mini-language
        # https://kuvapcsitrd01.kutztown.edu/~schwesin/fall20/csc223/lectures/Python_String_Formatting.html
        # https://en.wikipedia.org/wiki/Box-drawing_character
        width = 20
        if len(self.name) > 10:
            width = round(len(self.name) * 1.8)
            if width % 2 != 0:
                width -= 1
        sections = {
            "blank": f"│{width * ' '}│",
            "top": f"┌{width * '─'}┐",
            "bottom": f"└{width * '─'}┘",
            "sep": f"├{width * '─'}┤",
        }
        Chalk.set("bold")
        print(sections["top"])
        print(f"│{str(self.mass).ljust(width//2)}{str(self.number).rjust(width//2)}│")
        print(sections["blank"])
        print(f"│{str(self.symbol).center(width)}│")
        print(sections["blank"])
        print(f"│{str(self.name).center(width)}│")
        print(sections["bottom"])
        Chalk.reset()
        return

    def details(self):
        """
        Display the details of the element.
        """
        for key in self.data:
            print(f"{pydash.separator_case(key, ' ').title()}: {self.data[key]}")
        return

    def __str__(self):
        return f"{self.symbol}"

    def __repr__(self):
        return f"Element({self.symbol}, {self.name}, {self.number}, {self.mass})"

    def __eq__(self, other):
        if isinstance(other, Element):
            return self.symbol == other.symbol
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __add__(self, other):
        if isinstance(other, Element):
            return Formula({self: 1, other: 1})
        return NotImplemented

    def __sub__(self, other):
        return ValueError("Cannot subtract elements")

    def __mul__(self, other):
        if isinstance(other, int):
            return Formula({self: other})
        return NotImplemented

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        return ValueError("Cannot divide elements")

    def __rtruediv__(self, other):
        return ValueError("Cannot divide elements")

    def __floordiv__(self, other):
        return ValueError("Cannot divide elements")

    def __rfloordiv__(self, other):
        return ValueError("Cannot divide elements")

    def __mod__(self, other):
        return ValueError("Cannot mod elements")

    def __rmod__(self, other):
        return ValueError("Cannot mod elements")

    def __pow__(self, other):
        return ValueError("Cannot raise elements to a power")

    def __rpow__(self, other):
        return ValueError("Cannot raise elements to a power")

    def __lt__(self, other):
        if isinstance(other, Element):
            return self.number < other.number
        return NotImplemented

    def __le__(self, other):
        if isinstance(other, Element):
            return self.number <= other.number
        return NotImplemented

    def __gt__(self, other):
        if isinstance(other, Element):
            return self.number > other.number
        return NotImplemented

    def __ge__(self, other):
        if isinstance(other, Element):
            return self.number >= other.number
        return NotImplemented

    def __hash__(self):
        return hash(self.symbol)

    def __len__(self):
        return None

    def __getitem__(self, key):
        return self.data[key]

    def __iter__(self):
        return iter(self.data)


class Formula:
    """
    A class to represent a chemical formula.

    Args:
        elements (str, dict, Element): The elements in the formula.

    Attributes:
        elements (dict): The elements in the formula.
        mass (float): The molar mass of the formula.
        name (str): The name of the formula.
        count (int): The number of elements in the formula.

    Methods:
        display: Display the formula in a "periodic table"-like format.
        details: Display the details of the formula.
        __str__: Return the formula as a string.

    Examples:
        >>> print(Formula("H2O").mass)
        18.015
        >>> print(Formula("H2O").name)
        Water
    """

    def __init__(self, elements):
        if isinstance(elements, str):
            if elements.isdigit():
                raise ValueError("Invalid formula")
            elements = parse_formula(elements.strip())
        elif isinstance(elements, Element):
            elements = {elements: 1}
        else:
            if not isinstance(elements, dict):
                return ValueError("Invalid formula")
        self.elements = {
            Element(element).symbol: elements[element] for element in elements
        }
        self.mass = get_molar_mass(self)
        self.name = get_formula_name(reconstruct_formula(self.elements))
        if self.name == None:
            if len(self.elements) == 1:
                element = list(self.elements.keys())[0]
                if element in PERIODIC_TABLE["Symbol"].values:
                    self.name = Element(element).name
            else:
                self.name = "Unknown"
        else:
            self.name = str(self.name).title()
        self.count = 0
        for element in self.elements:
            self.count += self.elements[element]

    def __str__(self):
        return reconstruct_formula(self.elements)

    def __repr__(self):
        return f"Formula({self.elements})"

    def __eq__(self, other):
        if isinstance(other, Formula):
            return self.elements == other.elements
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __add__(self, other):
        if isinstance(other, Formula):
            return Formula({**self.elements, **other.elements})
        elif isinstance(other, Element):
            for element in self.elements:
                if element == other:
                    return Formula(
                        {**self.elements, **{element: self.elements[element] + 1}}
                    )
            return Formula({**self.elements, **{other: 1}})
        return NotImplemented

    def __sub__(self, other):
        if isinstance(other, Formula):
            return Formula(
                {
                    **self.elements,
                    **{
                        element: -1 * other.elements[element]
                        for element in other.elements
                    },
                }
            )
        elif isinstance(other, Element):
            for element in self.elements:
                if element == other:
                    return Formula(
                        {**self.elements, **{element: self.elements[element] - 1}}
                    )
            return ValueError("Cannot subtract element not in formula")
        return NotImplemented

    def __lt__(self, other):
        return self.mass < other.mass

    def __le__(self, other):
        return self.mass <= other.mass

    def __gt__(self, other):
        return self.mass > other.mass

    def __ge__(self, other):
        return self.mass >= other.mass

    def __hash__(self):
        return hash(self.elements)

    def __len__(self):
        return len(self.elements)

    def __getitem__(self, key):
        return self.elements[key]

    def __iter__(self):
        return iter(self.elements)

    def __contains__(self, item):
        return item in self.elements


def iselement(value):
    """
    Check if a value is an element.

    Args:
        value (any): The value to check.

    Returns:
        bool: True if the value is an element, False otherwise.
    """
    try:
        Element(value)
        return True
    except:
        return False


def isformula(value):
    """
    Check if a value is a formula.

    Args:
        value (any): The value to check.

    Returns:
        bool: True if the value is a formula, False otherwise.
    """
    try:
        Formula(value)
        return True
    except:
        return False


def mass_to_moles(molecule, mass):
    """
    Convert mass to moles.

    Args:
        molecule (Element, Formula, str): The molecule.
        mass (float): The mass.

    Returns:
        float: The moles.
    """
    if isinstance(molecule, Element) or isinstance(molecule, Formula):
        return mass / molecule.mass
    elif isinstance(molecule, str):
        return mass / Formula(molecule).mass
    return None


def moles_to_mass(molecule, moles):
    """
    Convert moles to mass.

    Args:
        molecule (Element, Formula, str): The molecule.
        moles (float): The moles.

    Returns:
        float: The mass.
    """
    if isinstance(molecule, Element) or isinstance(molecule, Formula):
        return moles * molecule.mass
    elif isinstance(molecule, str):
        return moles * Formula(molecule).mass
    return None


def mass_to_atoms(molecule, mass):
    """
    Convert mass to atoms.

    Args:
        molecule (Element, Formula, str): The molecule.
        mass (float): The mass.

    Returns:
        float: The atoms.
    """
    if isinstance(molecule, Element) or isinstance(molecule, Formula):
        return mass / molecule.mass * AVOGADRO * molecule.count
    elif isinstance(molecule, str):
        return mass / Formula(molecule).mass * AVOGADRO * Formula(molecule).count
    return None


def atoms_to_mass(molecule, atoms):
    """
    Convert atoms to mass.

    Args:
        molecule (Element, Formula, str): The molecule.
        atoms (float): The atoms.

    Returns:
        float: The mass.
    """
    if isinstance(molecule, Element) or isinstance(molecule, Formula):
        return atoms / AVOGADRO * molecule.mass
    elif isinstance(molecule, str):
        return atoms / AVOGADRO * Formula(molecule).mass
    return None


def mass_to_particles(molecule, mass):
    """
    Convert mass to particles.

    Args:
        molecule (Element, Formula, str): The molecule.
        mass (float): The mass.

    Returns:
        float: The particles.
    """
    if isinstance(molecule, Element) or isinstance(molecule, Formula):
        return mass / molecule.mass * AVOGADRO
    elif isinstance(molecule, str):
        return mass / Formula(molecule).mass * AVOGADRO
    return None


def particles_to_mass(molecule, particles):
    """
    Convert particles to mass.

    Args:
        molecule (Element, Formula, str): The molecule.
        particles (float): The particles.

    Returns:
        float: The mass.
    """
    if isinstance(molecule, Element) or isinstance(molecule, Formula):
        return particles / AVOGADRO * molecule.mass
    elif isinstance(molecule, str):
        return particles / AVOGADRO * Formula(molecule).mass
    return None


def moles_to_atoms(molecule, moles):
    """
    Convert moles to atoms.

    Args:
        molecule (Element, Formula, str): The molecule.
        moles (float): The moles.

    Returns:
        float: The atoms.
    """
    if isinstance(molecule, Element) or isinstance(molecule, Formula):
        return moles * AVOGADRO * molecule.count
    elif isinstance(molecule, str):
        return moles * AVOGADRO * Formula(molecule).count
    return None


def atoms_to_moles(molecule, atoms):
    """
    Convert atoms to moles.

    Args:
        molecule (Element, Formula, str): The molecule.
        atoms (float): The atoms.

    Returns:
        float: The moles.
    """
    if isinstance(molecule, Element) or isinstance(molecule, Formula):
        return atoms / AVOGADRO / molecule.count
    elif isinstance(molecule, str):
        return atoms / AVOGADRO / Formula(molecule).count
    return None


def get_empirical_formula(percentages):
    """
    Get the empirical formula from percentages of elements.

    Args:
        percentages (dict): The percentages of elements in the formula, summed to 1 or 100 (e.g. {'C': 0.5, 'H': 0.5} or {'C': 50, 'H': 50}).

    Returns:
        Formula: The empirical formula.
    """
    if isinstance(percentages, dict):
        elements = [key for key in percentages]
        if not all([iselement(element) for element in elements]):
            return ValueError("Invalid percentages (not all elements)")
        if sum(percentages.values()) == 1 or sum(percentages.values()) == 100:
            if sum(percentages.values()) == 100:
                percentages = {key: value / 100 for key, value in percentages.items()}
            empirical_num_atoms = {
                key: value / Element(key).mass for key, value in percentages.items()
            }
            empirical_ratio = {
                key: round(value / min(empirical_num_atoms.values()))
                for key, value in empirical_num_atoms.items()
            }
            for element in empirical_ratio:
                print(f"{element}: {empirical_ratio[element]}")
            return Formula(empirical_ratio)
    else:
        return ValueError("Invalid input (not a dictionary)")


def get_molecular_formula(empirical_formula, mass):
    """
    Get the molecular formula from the empirical formula and the mass.

    Args:
        empirical_formula (Formula, dict): The empirical formula.
        mass (float): The mass.

    Returns:
        Formula: The molecular formula.
    """
    if isinstance(empirical_formula, dict):
        empirical_formula = Formula(empirical_formula)
    if isinstance(empirical_formula, Formula):
        if isinstance(mass, (int, float)):
            molecular_mass = empirical_formula.mass
            molecular_ratio = {
                key: round((value * mass) / molecular_mass)
                for key, value in empirical_formula.elements.items()
            }
            return Formula(molecular_ratio)
    else:
        return ValueError("Invalid input (not a formula)")


def get_percent_composition(formula, round_result=False):
    """
    Get the percent composition of a formula.

    Args:
        formula (Formula): The formula.
        round_result (bool, optional): Whether to round the result. Defaults to False.

    Returns:
        dict: The percent composition of each element in the formula.
    """
    if isinstance(formula, Formula):
        if round_result:
            return {
                key: round((Element(key).mass * value / formula.mass) * 100, 2)
                for key, value in formula.elements.items()
            }
        return {
            key: (Element(key).mass * value / formula.mass) * 100
            for key, value in formula.elements.items()
        }
    else:
        return ValueError("Invalid input (not a formula)")


def get_percent_error(actual, theoretical):
    """
    Get the percent error between two numbers.

    Args:
        actual (int, float): The actual value.
        theoretical (int, float): The theoretical value.

    Returns:
        float: The percent error.
    """
    if isinstance(actual, (int, float)) and isinstance(theoretical, (int, float)):
        return abs(actual - theoretical) / theoretical * 100
    else:
        return ValueError("Invalid input (not a number)")
